-[library "command.c" , header "command.h" ]


#define LCURLY '{'
#define RCURLY '}'

-[private] typedef struct entry entry_t;

-[private]  struct entry {
	int id;
	strtab_t fields;
	-[ tree members for entry_t ]
};

-[private] typedef struct {
	entry_t *root;
} cmdtab_t;

-[ instantiate tree cmdtab_t :
	scope=private,
	node type= entry_t,
	comparision= intcompare,
	prefix= cmdtab ]

-[private] int intcompare(entry_t *A,entry_t *B) {
	return A->id - B->id;
}

-[private] cmdtab_t commands;

arg_t *blt_cmdnew(arg_t *call,frame_t *frame) {
	char *ident= get_varname(call,0,"cmd.new");
	-[private] int lastid;
	entry_t *E;
	-[ptr/new E]
	-[ptr/zero E]
	E->id= lastid++;
	cmdtab_insert(&commands, E);
	char buf[20];
	sprintf(buf,"%d",E->id);
	var_set(ident, arg_new_text(strdup(buf)), frame);
	return 0;
}

arg_t *blt_cmdset(arg_t *call,frame_t *frame) {
	arg_t *ident= arglist(call,0);
	arg_t *body= arglist(call,1);
	if (!ident || !body)
		wrong_args("cmd.set");
	char *itext= arg_to_string(ident,frame);
	int id= atoi(itext);
	free(itext);

	entry_t key;
	key.id= id;
	entry_t *E= cmdtab_find(&commands,&key);
	if (!E)
		runtime_error("cmd.set", "can not find command");

	token_t *T;
	int state= 'I';
	int depth=0;
	char *fieldname;
	arg_t *value;
	-[ for each T in list (&body->tokens) , begin ]
		switch(state) {
		case 'I':
			if (T->kind!='W' || T->data[0]=='$')
				runtime_error("cmd.set", "invalid field specifier");
			fieldname= T->data;
			state= 'B';
			value= arg_new();
			break;
		case 'B':
			if (T->kind=='{') /* } */ {
				state= 'L';
				depth= 1;
			} else {
				expand_token_into(T,frame,value);
				state= 'I';
				setfield(fieldname, E, value);
			}
			break;
		case 'L':
			if (T->kind=='{') /*}*/ {
				depth++;
				expand_token_into(T,frame,value);
				break;
			} 
			/*{*/ if (T->kind=='}') {
				depth--;
				if (depth) {
					expand_token_into(T,frame,value);
				} else {
					state= 'I';
					setfield(fieldname, E, value);
				}
				break;
			}
			expand_token_into(T,frame,value);
			break;
		}
	-[ end ]
	return 0;
}

-[private] void setfield(char *name, entry_t *E, arg_t *V) {
	arg_t *current;
	current= strtab_find(&E->fields, name);
	if (current) {
		arg_destroy_deep(current);
		strtab_remove(&E->fields, name);
	}
	strtab_insert(&E->fields, name, V);
}

-[private] void appendfield(char *name, entry_t *E,arg_t *V) {
	arg_t *current;
	current= strtab_find(&E->fields,name);
	if (current) {
		token_t *T;
		-[ for each T in list (&V->tokens) , begin ]
			-[clist/remove T]
			-[clist/insert T before (&current->tokens) ]
		-[ end ]
	} else {
		setfield(name, E,V);
	}
}

-[private] void 
expand_token_into(token_t *T, frame_t *frame, arg_t *dst) {
	arg_t *V;
	if (T->kind!='W' || T->data[0]!='$') 
		goto copy;
	V= var_get(T->data, frame);
	if (!V) 
		goto copy;
	-[ for each T in list (&V->tokens) , begin ]
		token_t *C= token_copy_deep(T);
		-[clist/insert C before (&dst->tokens) ]
	-[ end ]
	return ;
copy:
	T= token_copy_deep(T);
	-[clist/insert T before (&dst->tokens) ]
}

-[ lheader "parser.h" ] -[ lheader "table.h" ] -[ lheader "functions.h" ]
-[ header "stdlib.h"] -[ header "string.h" ] -[ header "stdio.h" ]
-[ lheader "string.h" ] -[ lheader "print.h"]

arg_t *dump_commands(arg_t *call,frame_t *F) {
	/* CAUTION: this is destructive man. */
	entry_t *E;
	while((E=cmdtab_destroy(&commands))) {
		print_entry(E);
	}
	return 0;
}

-[private] void print_entry(entry_t *E) {
	char *key;
	printf("Command %d  ------\n", E->id);
	while((key=strtab_getleast(&E->fields))) {
		printf("[%s] ",key);
		arg_t *A=strtab_remove(&E->fields, key);
		print_arg(A);
	}
}

-[public] int command_get_one(strtab_t *R) {
	entry_t *E;
	E= cmdtab_destroy(&commands);
	if (!E)
		return 1;
	*R= E->fields;
	free(E);
	return 0;
}


void command_unused_functions_do_not_call() {
	cmdtab_init(0);
	cmdtab_remove(0,0);
}



arg_t *blt_cmdmodify(arg_t *call,frame_t *frame) {
	arg_t *ident= arglist(call,0);
	arg_t *body= arglist(call,1);
	token_t *T;

	if (!ident || !body)
		wrong_args("cmd.modify");

	-[ for each T in list (&ident->tokens) , begin ]
		arg_t *V;
		token_t *T2;
		if (T->kind!='W')
			continue;
		if (T->data[0]!='$') {
			modify_command(T->data, body, frame);
			continue; 
		}
		V= var_get(T->data, frame);
		if (!V)
			exit(fprintf(stderr,
				"jj: (cmd.modify) can not find variable %s\n", T->data));
		-[ for each T2 in list (&V->tokens) , begin ]
			if (T2->kind=='W')
				modify_command(T2->data,body,frame);
		-[ end ]
	-[ end ]
	return 0;
}

-[private] void 
modify_command(char *target,arg_t *body, frame_t *frame) {
	entry_t *E= find_by_target(commands.root,target);
	if (!E) {
		fprintf(stderr,"can not find command that creates %s\n",
			target);
		exit(1);
	}

	token_t *T;
	int state= 'I';
	int depth=0;
	char *fieldname;
	arg_t *value;
	int modtype;

	-[ for each T in list (&body->tokens) , begin ]
		switch(state) {
		case 'I':
			if (T->kind!='W') 
				runtime_error("cmd.modify", "invalid modification");
			modtype= T->data[0];
			if (modtype!='+' && modtype!='=')
				runtime_error("cmd.modify", "invalid modification operator");
			state= 'F';
			break;
		case 'F':
			if (T->kind!='W') 
				runtime_error("cmd.modify", "invalid field specification");
			fieldname= T->data;
			state= 'B';
			value= arg_new();
			break;
		case 'B':
			if (T->kind==LCURLY)  {
				state= 'L';
				depth= 1;
			} else {
				expand_token_into(T,frame,value);
				state= 'I';
				if (modtype=='=')
					setfield(fieldname, E, value);
				else
					appendfield(fieldname, E, value);
			}
			break;
		case 'L':
			if (T->kind==LCURLY)  {
				depth++;
				expand_token_into(T,frame,value);
				break;
			} 
			 if (T->kind==RCURLY) {
				depth--;
				if (depth) {
					expand_token_into(T,frame,value);
				} else {
					state= 'I';
					if (modtype=='=')
						setfield(fieldname, E, value);
					else
						appendfield(fieldname,E,value);
				}
				break;
			}
			expand_token_into(T,frame,value);
			break;
		}
	-[ end ]
}

-[private] entry_t *find_by_target(entry_t *node, char *name) {
	arg_t *out;
	entry_t *e=0;
	if (!node)
		return 0;
	out= strtab_find(&node->fields, "outputs");
	if (out) {
		token_t *T;
		-[ for each T in list (&out->tokens) , begin ]
			if (!strcmp(T->data, name))
				return node;
		-[ end ]
	}
	if (node->left) e= find_by_target(node->left, name);
	if (!e && node->right) e=find_by_target(node->right, name);
	return e;
}

