
#include <string.h>
#include <stdlib.h>


#include "table.h"
 typedef struct entry_s entry_t; struct entry_s {
	char *string;
	void *data;
	
  entry_t *left, *right;
  int height;

}; typedef struct {
	entry_t *root;
} strtable_t;
 static  void
strtree_init(strtable_t *t) ; static  void*
strtree_find(strtable_t *t,entry_t *key) ; static  int
strtree_insert(strtable_t *t,entry_t *node) ; static  entry_t*
strtree_insert_node(entry_t *t,entry_t *n) ; static  entry_t*
strtree_balance(entry_t *t) ; static  entry_t*
strtree_remove(strtable_t *t, entry_t *key) ; static  entry_t*
strtree_remove_node(entry_t *t,entry_t *node) ; static  entry_t*
strtree_rotate_right(entry_t *g) ; static  entry_t*
strtree_rotate_left(entry_t *g) ; static  entry_t*
strtree_remove_max(entry_t *t,entry_t **pl) ; static  entry_t*
strtree_remove_min(entry_t *t,entry_t **pl) ; static  void
strtree_calc_height(entry_t *t) ; static  entry_t*
strtree_destroy(strtable_t *t) ; static  entry_t*
strtree_destroy_one_node(entry_t *node,entry_t** plink) ; static  int strcompare(entry_t *A,entry_t *B) ; static  char *least(entry_t *E) ;








 






 static  void
strtree_init(strtable_t *t) { t->root= 0; }


 static  void*
strtree_find(strtable_t *t,entry_t *key) {
	entry_t *n= t->root;
	while(n) {
		int u;
		u= strcompare(key,n);
		if (u==0) return n;
		else if (u<0) n=n->left;
		else n=n->right;
	}
	return 0;
}

 static  int
strtree_insert(strtable_t *t,entry_t *node) {
	if (strtree_find(t,node)) {
		return 1;
	} else {
		node->height= 1;
		node->left= node->right= 0;
		t->root= strtree_insert_node(t->root,node);
		return 0;
	}
}

 static  entry_t*
strtree_insert_node(entry_t *t,entry_t *n) {
	int u;
	if (!t)
		return n;
	u= strcompare(n,t);
	if (u<0)
		t->left= strtree_insert_node(t->left,n);
	else
		t->right= strtree_insert_node(t->right,n);
	return strtree_balance(t);
}

#ifndef TREE_height
#define TREE_height(x) ((x)? (x)->height : 0)
#endif

 static  entry_t*
strtree_balance(entry_t *t) {
	int u;
	strtree_calc_height(t);
	u= TREE_height(t->left)-TREE_height(t->right);
	if (u>1)
		return strtree_rotate_right(t);
	else if (u<-1)
		return strtree_rotate_left(t);
	else return t;
}

 static  entry_t*
strtree_remove(strtable_t *t, entry_t *key) {
	entry_t *node;
	node= strtree_find(t,key);
	if (node)
		t->root= strtree_remove_node(t->root,node);
	return node;
}

 static  entry_t*
strtree_remove_node(entry_t *t,entry_t *node) {
	int u;
	u= strcompare(node,t);
	if (u<0) 
		t->left= strtree_remove_node(t->left,node);
	else if (u>0)
		t->right= strtree_remove_node(t->right,node);
	else {
		if (!t->right || !t->left) {
			if (t->right)
				return t->right;
			else
				return t->left;
		} else {
			entry_t *p;
			if (t->left->height>t->right->height)
				p= strtree_remove_max(t->left, &t->left);
			else 
				p= strtree_remove_min(t->right, &t->right);
			p->left= t->left;
			p->right= t->right;
			p->height=1;
			t= p;
		}
	}
	return strtree_balance(t);
}

 static  entry_t*
strtree_rotate_right(entry_t *g) {
	entry_t *f,*e;
	f= g->left;
	e= f->right;
	f->right= g;
	g->left= e;
	strtree_calc_height(g);
	strtree_calc_height(f);
	return f;
}

 static  entry_t*
strtree_rotate_left(entry_t *g) {
	entry_t *f,*e;
	f= g->right;
	e= f->left;
	f->left= g;
	g->right= e;
	strtree_calc_height(g);
	strtree_calc_height(f);
	return f;
}


 static  entry_t*
strtree_remove_max(entry_t *t,entry_t **pl) {
	if (t->right) {
		entry_t *z;
		z= strtree_remove_max(t->right,&t->right);
		strtree_calc_height(t);
		return z;
	}
	*pl= t->left;
	return t;
}

 static  entry_t*
strtree_remove_min(entry_t *t,entry_t **pl) {
	if (t->left) {
		entry_t *z;
		z= strtree_remove_min(t->left,&t->left);
		strtree_calc_height(t);
		return z;
	}
	*pl= t->right;
	return t;
}

 static  void
strtree_calc_height(entry_t *t) {
	int a,b;
	a= TREE_height(t->left)+1;
	b= TREE_height(t->right)+1;
	if (a>b) t->height=a;
	else t->height= b;
}

 static  entry_t*
strtree_destroy(strtable_t *t) {
	if (!t->root)
		return 0;
	return strtree_destroy_one_node(t->root,&t->root);
}

 static  entry_t*
strtree_destroy_one_node(entry_t *node,entry_t** plink) {
	if (node->left)
		return strtree_destroy_one_node(node->left, &node->left);
	else if (node->right)
		return strtree_destroy_one_node(node->right, &node->right);
	*plink= 0;
	return node;
}



#define getT   strtable_t *T= (strtable_t*) _T

 void* strtab_find(strtab_t *_T, char *key) {
	getT;
	entry_t E;
	E.string= key;
	entry_t *R;
	R= strtree_find(T,&E);
	if (R)
		return R->data;
	else
		return 0;
}

 static  int strcompare(entry_t *A,entry_t *B) {
	return strcmp(A->string,B->string);
}



 void strtab_insert(strtab_t *_T, char *key, void *data) {
	getT;
	entry_t *E;
	
	(E)= (typeof(E)) malloc(sizeof(*(E)));


	E->string= strdup(key);
	E->data= data;
	strtree_insert(T,E);
}




 strtab_t *strtab_new() {
	strtable_t *T;
	
	(T)= (typeof(T)) malloc(sizeof(*(T)));


	
	memset( (T), 0, sizeof(*(T)));
	

	return (strtab_t*) T;
}

 void*
strtab_destroy(strtab_t *_T) {
	getT;
	entry_t *E;
	E= strtree_destroy(T);
	if (E) {
		void *data= E->data;
		free(E->string);
		free(E);
		return data;
	} else 
		return 0;
}

 void*
strtab_remove(strtab_t *_T,char *key) {
	getT;
	entry_t E,*R;
	E.string= key;
	R= strtree_remove(T, &E);
	if (R) {
		void *data= R->data;
		free(R->string);
		free(R);
		return data;
	} else
		return 0;
}


 char *strtab_getleast(strtab_t *_T) {
	getT;
	return least(T->root);
}

 static  char *least(entry_t *E) {
	if (!E)
		return 0;
	if (E->left)
		return least(E->left);
	return E->string;
}


void strtab_unused_functions_do_not_call() {
	strtree_init(0);
}
