-[ library "parser.c", header "parser.h" ]

![flex "-L" ]
%option stdout
%option noyywrap
%option yylineno
NS [^\001-\040{}]
%%
"{"  { return '{'; }
"}"  { return '}'; }
{NS}+ { return 'W'; }
<<EOF>> {return 'E'; }
. { return 'S'; }
\n { }
%%
![end]

-[private] int false;

-[private] char* parse_stream(parser_t *P, token_t *T) {
	switch(T->kind) {
	case '{':
		if (P->arg) {
			T= token_copy_deep(T);
			-[ clist/insert T before (&P->arg->tokens) ]
			P->depth++;
		} else {
			arg_t *A;
			-[ ptr/new A ]
			-[ clist/init (&A->tokens) ]
			-[ clist/insert A before (&P->stmt->args) ]
			P->arg= A;
			P->depth= 1;
		}
		break;
	case '}':
		if (!P->arg) {	   				
			char buf[200];
			sprintf(buf,"%d: unmatched '}'", yylineno);
			return strdup(buf);
		}
		if (P->depth==1) {
			P->arg= 0;
		} else {
			T= token_copy_deep(T);
			-[ clist/insert T before (&P->arg->tokens) ]
			P->depth--;
		}
		break;
	case 'W':
		if (P->arg) {
			T= token_copy_deep(T);
			-[clist/insert T before (&P->arg->tokens) ]
		} else {
			P->stmt->func= strdup(T->data);
			push_statement(P);
		}
		break;
	case 'E':
		break;
	}
   	return 0;
}

-[ header "string.h" ]

-[public] char *parse_file(char *fn,stmt_t **result) {
	-[ open and switch to the file ]
	parser_t P;
	char *err=0;
	parser_init(&P);
	token_t T;
	while(!err) {
		char *data= 0;
		int t= yylex();
		if (t=='E')
			break;
		switch(t) {
			case 'W': data= strdup(yytext);
			case '{':
			case '}': 
				T.kind= t;
				T.lineno= yylineno;
				T.data= data;
				err= parse_stream(&P,&T);
				break;
		}
	}
	if (!err)
		err= parser_finish(&P);
	-[close file]
	stmt_t *last= P.stmts->prev;
	-[clist/remove last]
	*result= P.stmts;
	return err;
}


=[ open and switch to the file ]
	FILE *f;
	if (strcmp(fn,"-"))
		f= fopen(fn,"rb");
	else
		f= stdin;
	-[ header "stdio.h" ]
	if (!f) {
		char buf[2000];
		sprintf(buf,"can not open file %s: %s", fn, strerror(errno));
		-[ header "errno.h" ]
		return strdup(buf);
	}
	yyrestart(f);

=[ close file ]
	fclose(f);
=[]

-[private] char* parser_finish(parser_t *P) {
	if (P->arg)
		return "unexpected end of file";
	return 0;
}
		

-[private] void push_statement(parser_t *P) {
	stmt_t *I;
  	-[ptr/new I]
	-[ ptr/zero I ]
	I->args.prev= I->args.next= &I->args;
	-[clist/insert I before (P->stmts) ]
	P->stmt= I;
}

-[private] void
parser_init(parser_t *P) {
	stmt_t *I;
	P->arg= 0;
	-[ptr/new I]
	-[ptr/zero I]
	-[clist/init I]
	P->stmts= I;
	push_statement(P);
}

-[public] typedef struct token_t {
	struct token_t *prev,*next;
	int lineno;
	int kind;
	char *data;
} token_t;

-[public] typedef struct arg_t {
	token_t tokens;
	struct arg_t *prev,*next;
} arg_t;

-[public] typedef struct stmt_t { 
	char *func;
	struct arg_t args;
	struct stmt_t *prev,*next;
} stmt_t;

-[private] typedef struct {
	stmt_t *stmt;
	arg_t *arg;
	int depth;
	stmt_t *stmts;
} parser_t;



-[public] char *parse_arg(arg_t *stream,stmt_t **result) {
	parser_t P;
	token_t *t;
	char *err=0;
	parser_init(&P);
	-[ for each t in list (&stream->tokens) , begin ]
		err= parse_stream(&P,t);
		if (err)
			break;
	-[ end ]
	if (!err)
		err= parser_finish(&P);
	stmt_t *last= P.stmts->prev;
	-[clist/remove last]
	*result= P.stmts;
	return err;
}


-[public] arg_t *arg_copy_deep(arg_t *orig) {
	arg_t *na= arg_new();
	token_t *t;
	-[ for each t in list (&orig->tokens) , begin ]
		token_t *c;
		c= token_copy_deep(t);
		-[clist/insert c before (&na->tokens) ]
	-[ end ]
	return na;
}

-[public] token_t *token_copy_deep(token_t *orig) {
	token_t *t;
	-[ptr/new t]
	*t= *orig;
	if (t->kind=='W')
		t->data= strdup(t->data);
	return t;
}

-[public] void arg_destroy_deep(arg_t *a) {
	token_t *T;
	-[ for each T in list (&a->tokens) , begin ]
		-[ clist/remove T ]
		token_destroy_deep(T);
	-[ end ]
	free(a);
}

-[public] void token_destroy_deep(token_t *T) {
	if (T->kind=='W')
		free(T->data);
	free(T);
}

-[public] arg_t* arglist_copy_deep(arg_t *orig) {
	arg_t *nl= arg_new();
	arg_t *arg;
	-[for each arg in list orig, begin ]
		arg_t *copy;
		copy= arg_copy_deep(arg);
		-[clist/insert copy before nl]
	-[ end ]

	return nl;
}

-[public] stmt_t* stmt_copy_deep(stmt_t* orig) {
	stmt_t *ns;
	-[ptr/new ns]
	-[ptr/zero ns]
	ns->func= strdup(orig->func);
	ns->args.prev= ns->args.next= &ns->args;

	arg_t *arg;
	-[for each arg in list (&orig->args), begin ]
		arg_t *copy;
		copy= arg_copy_deep(arg);
		-[clist/insert copy before (&ns->args)]
	-[ end ]
	return ns;
}

-[public] stmt_t *stmtlist_copy_deep(stmt_t *orig) {
	stmt_t *ns;
	-[ptr/new ns]
	-[ptr/zero ns]
	ns->prev= ns->next= ns;

	stmt_t *S;
	-[for each S in list orig, begin ]
		stmt_t *copy= stmt_copy_deep(S);
		-[clist/insert copy before ns]
	-[ end ]
	return ns;
}

-[public] void stmtlist_destroy_deep(stmt_t *L) {
	stmt_t *S;
	-[ for each S in list L, begin ]
		stmt_destroy_deep(S);
	-[ end ]
	free(S);
}

-[public] void arglist_destroy_deep(arg_t *L) {
	arg_t *S;
	-[ for each S in list L, begin ]
		arg_destroy_deep(S);
	-[ end ]
	free(S);
}

-[public] void stmt_destroy_deep(stmt_t *S) {
	arg_t *A;
	free(S->func);
	-[ for each A in list (&S->args) , begin ]
		arg_destroy_deep(A);
	-[ end ]
	free(S);
}

-[public] arg_t *arg_new() {
	arg_t *A;
	-[ptr/new A]
	-[clist/init A]
	-[clist/init (&A->tokens)]
	return A;
}

-[public] token_t *token_new() {
	token_t *T;
	-[ptr/new T]
	-[ptr/zero T]
	-[clist/init T]
	return T;
}

-[public] arg_t *arg_new_text(char *txt) {
	arg_t *A= arg_new();
	token_t *T= token_new();
	T->kind= 'W';
	T->data= txt;
	T->lineno= 0;
	-[clist/insert T before (&A->tokens) ]
	return A;
}

void parser_unused_functions_dont_call() {
	yyunput(0,0);
}

