-[ library "string.c", header "string.h" ]


/* $X $Y str.cat */
arg_t* string_cat(arg_t *call, frame_t *F) {
	arg_t *args= flatten(call);

	char *result= 0;
	int len;
	
	len= stringcat(args, 0, F);
	result= malloc(len);
	stringcat(args, result, F);

	arg_t *a;
	token_t *T;
	-[ptr/new T]
	T->kind= 'W';
	/* FIXME: args can be empty, so no
		guarantee that this will be a logical value */
	T->lineno= args->tokens.next->lineno;
	T->data= result;

	-[ ptr/new a]
	a->tokens.prev= a->tokens.next= &a->tokens;
	-[ clist/insert T before (&a->tokens) ]

	arg_destroy_deep(args);
	return a;
}

-[public] char *arg_to_string(arg_t *arg,frame_t *F) {
	char *result= 0;
	int len;
	
	len= stringcat(arg, 0, F);
	result= malloc(len);
	stringcat(arg, result, F);
	return result;
}

-[public] int stringcat(arg_t *arg,char *out,frame_t *F) {
	token_t *T;
	int len= 0;
	-[ for each T in list (&arg->tokens), begin ]
		if (T->kind!='W')
			continue;
		if (T->data[0]=='$') {
			arg_t *a=0;
			token_t *subt;
			a= var_get(T->data, F);
			if (!a)
				exit(fprintf(stderr,"can not find variable %s\n",T->data));
			-[ for each subt in list (&a->tokens) , begin ]
				switch(subt->kind) {
				case '{': if (out) out[len]= '{'; len++; break;
				case '}': if (out) out[len]= '}'; len++; break;
				case 'W': if (out)
						strcpy(out+len,subt->data);
					len+= strlen(subt->data);
					break;
				}
			-[ end ]
		} else {
			if (out)
				strcpy(out+len, T->data);
			len+= strlen(T->data);
		}
	-[ end ]
	if (out)
		out[len]= 0;
	return len+1;
}



-[ header "stdlib.h" ]
-[ lheader "parser.h" ]
-[ lheader "table.h" ]
-[ lheader "functions.h" ]
-[ header "stdio.h" ]
-[ lheader "extlibs.h" ]
-[ header "stdarg.h" ]


-[ private ] arg_t *flatten(arg_t *orig) {
	arg_t *R;
	-[ptr/new R]
	R->tokens.prev= R->tokens.next= &R->tokens;
	arg_t *A;
	-[ for each A in list orig, begin ]
		token_t *T;
		-[ for each T in list (&A->tokens) , begin ]
			token_t *C;
			C= token_copy_deep(T);
			-[clist/insert C before (&R->tokens) ]
		-[ end ]
	-[ end ]
	return R;
}

arg_t *blt_strsfxrep(arg_t *call,frame_t *frame) {
	arg_t *input = arglist(call,0);
	arg_t *oldsfx_a= arglist(call,1);
	arg_t *newsfx_a= arglist(call,2);
	arg_t *output_a= arglist(call,3);

	if (!input || !oldsfx_a || !newsfx_a || !output_a)
		wrong_args("rep.sfx");

	token_t *output= toklist(&output_a->tokens,0);
	if (!output || output->kind!='W' || output->data[0]!='$')
		wrong_args("rep.sfx");

	token_t *oldsfx= toklist(&oldsfx_a->tokens,0);
	if (!oldsfx || oldsfx->kind!='W')
		wrong_args("rep.sfx");

	token_t *newsfx= toklist(&newsfx_a->tokens,0);
	if (!newsfx || newsfx->kind!='W')
		wrong_args("rep.sfx");

	arg_t *nV;
	-[ptr/new nV]
	nV->tokens.prev= nV->tokens.next= &nV->tokens;

	token_t *oV;
	-[ for each oV in list (&input->tokens) , begin ]
		if (oV->kind!='W')
			continue;
		if (oV->data[0]!='$') {
			token_t *T= sfxrep( oV, oldsfx->data, newsfx->data);
			-[ clist/insert T before (&nV->tokens) ]
		} else {
			arg_t *A;
			A= var_get(oV->data, frame);
			if (!A) {
				token_t *T= token_copy_deep(oV);
				-[clist/insert T before (&nV->tokens) ]
			} else {
				token_t *S;
				-[ for each S in list (&A->tokens) , begin ]
					token_t *T= sfxrep(S, oldsfx->data,  newsfx->data);
					-[clist/insert T before (&nV->tokens) ]
				-[ end ]
			}
				
		}
	-[ end ]

	var_set(output->data, nV,frame);


	return 0;
}


-[private] token_t*
sfxrep(token_t *srct,char *orig,char *rep) {
	char *V;
	char *src= srct->data;
	int oL= strlen(orig);
	int rL= strlen(rep);
	int sL= strlen(src);

	if (sL<oL || strcmp(src+sL-oL,orig))
		V= strdup(src);
	else {
		V= malloc(sL-oL+rL+1);
		memcpy(V, src, sL-oL);
		memcpy(V+sL-oL, rep, rL);
		V[sL-oL+rL]= 0;
	}
	
	token_t *T;
	-[ptr/new T]
	T->kind= 'W';
	T->data= V;
	T->lineno= srct->lineno;
	return T;
}

-[ header "string.h" ]

-[public] char *
arg_to_string_noforce(arg_t *arg,frame_t *F) {
	char *result;
	int len;
	len= prep_string(arg, F,0);
	if (len<2)
		return 0;
	result= malloc(len);
	prep_string(arg,F,result);
	result[strlen(result)-1]= 0;
	return result;
}

-[private] int 
prep_string(arg_t *arg, frame_t *frame, char *result) {
	token_t *T;
	int len=0;
	-[ for each T in list (&arg->tokens), begin ]
		if (T->kind!='W')  {
			len= put_token(result,len,T->kind);
		} else if (T->data[0]!='$') {
			len= put_string( result, len, T->data );
		} else {
			arg_t *A;
			A= var_get(T->data, frame);
			if (A) {
				token_t *S;
				-[ for each S in list (&A->tokens) , begin ]
					if (S->kind=='W')
						len= put_string(result,len,S->data);
					else
						len= put_token(result,len,S->kind);
				-[ end ]
			} else {
				len= put_string( result, len, T->data );
			}
		}
	-[ end ]
	if (result)
		result[len]= 0;
	return len+1;
}

-[ private] int put_string(char *result, int len, char *src) {
	int L= strlen(src);
	if (result) {
		memcpy(result+len, src, L);
		result[len+L]= ' ';
	}
	len+= L+1;
	return len;
}

-[private] int put_token(char *result, int len, int tk) {
	char buf[2];
	buf[0]= tk;
	buf[1]= 0;
	return put_string(result, len, buf);
}


arg_t *strprepend(arg_t *call,frame_t *F) {
	char *varname;
	arg_t *pfx;
	arg_t *lst;
	arg_t *value;
	char *pfxv;
	token_t *T;

	varname= get_varname(call, 0, "str.prepend");
	pfx= arglist(call,1);
	lst= arglist(call,2);

	value= arg_new();

	if (!pfx || !lst)
		wrong_args("str.prepend");
	pfxv= arg_to_string(pfx, F);

	-[ for each T in list (&lst->tokens) , begin ]
		token_t *T3;
		arg_t *varval;
		if (T->kind!='W') {
			token_t *T2= token_copy_deep(T);
			-[ clist/insert T2 before (&value->tokens) ]
			continue;
		}
		if (T->data[0]!='$') {
			prepend_string_into_arg( pfxv, T->data, value, T->lineno);
			continue; 
		}
		varval= var_get(T->data, F);
		if (!varval)
			runtime_error("str.prepend", "can not find variable");
		-[ for each T3 in list (&varval->tokens), begin ]
			if (T3->kind=='W') 
				prepend_string_into_arg(pfxv, T3->data, value, T->lineno);
		-[ end ]
	-[ end ]
	var_set(varname, value, F);
	return 0;
}

-[private] void prepend_string_into_arg
	(char *pfx,char *sfx, arg_t *a,int lno)
{
	token_t *T2;
	char *v;
	v= mprintf("%s%s",pfx,sfx);
	T2= token_new();
	T2->kind= 'W';
	T2->data= v;
	T2->lineno= lno;
	-[clist/insert T2 before (&a->tokens) ]
}
