-[ c library "schedule.c", header "schedule.h" ]

#define MAXUNSIGNED 0xffffffffu

-[public] void
run_commands() {
	int change=1;
	create_command_and_output_list();
	while(cmds.prev!=&cmds && change) {
		cmdlist_t *c;
		change= 0;
		-[ for each c in list (&cmds) , begin ]
				if (!command_is_ready(&c->fields))
					continue;
				execute_command(&c->fields);
				-[clist/remove c]
				cmd_destroy(&c->fields);
				free(c);
				change=1;
		-[ end ]
	}
	if (cmds.prev!=&cmds) {
		cmdlist_t *c;
		fprintf(stderr,"jj: there are one or more cyclic dependencies\n");
		fprintf(stderr,"jj: regarding the generation of following files:\n");
		-[ for each c in list (&cmds) , begin ]
			print_outputs(&c->fields);
		-[ end ]
		fprintf(stderr,"jj: I don't know what these cycles are\n");
		fprintf(stderr,"jj: but I'm almost pretty sure that\n");
		fprintf(stderr,"jj: they are there. \n");
		exit(1);
	}	
}

-[private] void print_outputs(strtab_t *c) {
	arg_t *out;
	token_t *T;
	out= strtab_find(c,"outputs");
	if (out) {
		fprintf(stderr,"jj: ");
		-[ for each T in list (&out->tokens) , begin ]
			fprintf(stderr, "%s ",T->data);
		-[ end ]
		fprintf(stderr,"\n");
	}
}


-[private] void
create_command_and_output_list() {
	while(1) {
		strtab_t cmd;
		cmdlist_t *c;
		arg_t *O;
		if (command_get_one(&cmd))
			return ;
		-[ptr/new c]
		-[ private] typedef struct cmdlist {
			struct cmdlist *prev,*next;
			strtab_t fields;
		} cmdlist_t;
		+[private variables] static cmdlist_t cmds = {&cmds, &cmds}; +[ end]
		c->fields= cmd;
		-[clist/insert c before (&cmds) ]
		-[ private] strtab_t outputs;
		O= strtab_find(&cmd,"outputs");
		if (!O)
			continue;
		token_t *T;
		-[ for each T in list (&O->tokens) , begin ]
			if (T->kind!='W') continue;
			if (!strtab_find(&outputs,T->data))
				strtab_insert(&outputs, T->data, (void*) 1);
		-[ end ]
	}
}

-[private] int command_is_ready(strtab_t *c) {
	token_t *T;
	arg_t *triggers;
	triggers= strtab_find(c, "triggers");
	if (triggers) {
		-[ for each T in list (&triggers->tokens) , begin ]
			if (T->kind!='W')
				continue;
			if (strtab_find(&outputs,T->data))
				return 0;
		-[ end ]
	}
	-[ gcc deps : check whether any trigger is in outputs ]
	return 1;
}

-[private] int command_is_clean(strtab_t *c) {
	unsigned min_o= MAXUNSIGNED,max_i=0;
	arg_t *A;
	token_t *T;

	A= strtab_find(c,"triggers");
	if (A) {
		-[for each T in list (&A->tokens) , begin ]
			if (T->kind!='W') continue;
			unsigned mt= modtime_input(T->data);
			if (mt>max_i)
				max_i= mt;
		-[ end ]
	}

	-[ gcc deps: process modtimes of triggers ]

	A= strtab_find(c, "outputs");
	if (A) {
		-[ for each T in list (&A->tokens) , begin ]
			if (T->kind!='W') continue;
			unsigned mt= modtime_output(T->data);
			if (mt<min_o)
				min_o= mt;
		-[ end ]
	}
	return max_i < min_o;
}

-[private] void cmd_destroy(strtab_t *C) {
	arg_t *A;
	return ;
	while((A=strtab_destroy(C)))
		arg_destroy_deep(A);
	free(C);
}

-[private] void
execute_command(strtab_t *c) {
	int output_modtimes=0;
	arg_t *out;
	token_t *T;
	if (!command_is_clean(c)) {
		char *action;
		char *message;

		action= prepare_string(c,"action");
		message= prepare_string(c,"message");

		-[ print the message ]
		if (option_norun) {
			output_modtimes= 'M';
		} else {
			if (system(action))
				exit(fprintf(stderr,"jj: child failed. exiting.\n"));
			output_modtimes= 'C';
		}
	}
	out= strtab_find(c,"outputs");
	if (out)
		-[ for each T in list (&out->tokens) , begin ]
				if (T->kind!='W')
					continue;
				if (output_modtimes=='M')
					modtime_force(T->data, MAXUNSIGNED);
				else if (output_modtimes=='C')
					modtime_update(T->data);
				strtab_remove(&outputs, T->data);
		-[ end ]
}

=[ print the message ]
	if (!option_noprefix)
		printf("jj: ");
	if (!action && !message) {
		printf("'Executing' a mystery command with no\n");
		if (!option_noprefix)
			printf("jj: ");
		printf("action or message in it..\n");
	} else {
		if (option_printmsg) {
			if (message)
				printf("%s\n",message);
			else
				printf("%s [NO MESSAGE DEFINED]\n", action);
		} else {
			if (action)
				printf("%s\n",action);
			else
				printf("%s [NO ACTION DEFINED]\n", message);
		}
	}
=[]

-[private] char*
prepare_string(strtab_t *c,char *field) {
	arg_t *A;
	char *result;
	int len;
	A= strtab_find(c, field);
	if (!A)
		return 0;
	len= prep_string(A, c, 0);
	if (len<2)
		return 0;
	result= malloc(len);
	prep_string(A,c,result);
	return result;
}


-[private] int 
prep_string(arg_t *arg, strtab_t *c, char *result) {
	token_t *T;
	int len=0;
	-[ for each T in list (&arg->tokens), begin ]
		if (T->kind!='W')  {
			len= put_token(result,len,T->kind);
		} else if (T->data[0]!='$') {
			len= put_string( result, len, T->data );
		} else {
			arg_t *A;
			A= strtab_find(c, T->data+1);
			if (A) {
				token_t *S;
				-[ for each S in list (&A->tokens) , begin ]
					if (S->kind=='W')
						len= put_string(result,len,S->data);
					else
						len= put_token(result,len,S->kind);
				-[ end ]
			} else {
				len= put_string( result, len, T->data );
			}
		}
	-[ end ]
	if (result)
		result[len]= 0;
	return len+1;
}

-[ private] int put_string(char *result, int len, char *src) {
	int L= strlen(src);
	if (result) {
		memcpy(result+len, src, L);
		result[len+L]= ' ';
	}
	len+= L+1;
	return len;
}

-[private] int put_token(char *result, int len, int tk) {
	char buf[2];
	buf[0]= tk;
	buf[1]= 0;
	return put_string(result, len, buf);
}

-[private] unsigned
modtime_input(char *fn) {
	unsigned mt;
	mt= modtime(fn);
	if (!mt)
		exit(fprintf(stderr,"can not stat %s: %s\n",fn,strerror(errno)));
	return mt;
}

/* This function is used for checking output files. 
   If the file doesn`t exist, it`s assumed to have a 
   very old timestamp, i.e. 0 == 1970 */
-[private] unsigned
modtime_output(char *fn) {
	return modtime(fn);
}

/* weak inputs are just like outputs. if they don`t exist,
   they simply cause a generation. hopefully, this generation
   will remove them from the weak input pool (i.e. outdated -MMD). */
-[private] unsigned
modtime_weak_input(char *fn) {
	return modtime_output(fn);
}

-[private] unsigned modtime(char *fn) {
	unsigned V;
	/* a cache is mandatory. sometimes we force
	modtimes on files and sometimes the file does not
	fucking exist in the first place */
	V= (unsigned) strtab_find(&modtimes,fn);
	if (V)
		return V;
	struct stat sbuf;
	if (stat(fn,&sbuf))
		return 0;
	V= sbuf.st_mtime;
	if (!V)
		V++;
	strtab_insert(&modtimes, fn, (void*) V);
	return V;
}

-[private] void modtime_force(char *fn,unsigned V) {
	if (strtab_find(&modtimes,fn))
		strtab_remove(&modtimes,fn);
	strtab_insert(&modtimes, fn, (void*) V);
}

-[private] strtab_t modtimes;

-[private] void modtime_update(char *fn) {
	unsigned V;
	V= (unsigned) strtab_find(&modtimes,fn);
	if (V)
		strtab_remove(&modtimes,fn);
	struct stat sbuf;
	if (stat(fn,&sbuf)) {
		strtab_insert(&modtimes,fn, (void*) MAXUNSIGNED);
		return ;
	}
	V= sbuf.st_mtime;
	if (!V)
		V++;
	strtab_insert(&modtimes, fn, (void*) V);
}

-[ lheader "parser.h" ] -[ lheader "table.h" ]
-[ lheader "main.h" ]
-[ header "unistd.h" ] -[header "stdlib.h"] -[ header "stdio.h" ]
-[ header "errno.h" ] -[ header "string.h"]
-[ lheader "gccdeps.h" ]
-[ header "sys/stat.h" ] -[lheader "command.h" ]


=[ gcc deps : check whether any trigger is in outputs ] {
	arg_t *out;
	out= strtab_find(c, "outputs");
	if (out) {
		int not_ready= 0;
		-[ for each T in list (&out->tokens) , begin ]
			char *gdeps;
			char *trigger;
			if (T->kind!='W') continue;
			gdeps= get_gcc_deps(T->data);
			if (!gdeps)
				continue;
			-[ for each trigger in string gdeps/' ', begin ]
				if (trigger) {
					if (strtab_find(&outputs,trigger)) {
						not_ready= 1;
						-[ break ]
					}
				}
			-[ end ]
			if (not_ready)
				break;
		-[ end ]
		if (not_ready)
			return 0;
	}
}


=[ gcc deps: process modtimes of triggers ] {
	A= strtab_find(c,"outputs");
	if (A) {
		-[for each T in list (&A->tokens) , begin ]
			char *gdeps;
			char *trigger;
			if (T->kind!='W') continue;
			gdeps= get_gcc_deps(T->data);
			if (!gdeps)
				continue;
			-[ for each trigger in string gdeps/' ', begin]
				if (trigger) {
					unsigned mt= modtime_weak_input(trigger);
					if (mt>max_i)
						max_i= mt;
				}
			-[ end ]
		-[ end ]
	}
}