-[ library "table.c", header "table.h" ]

-[private] typedef struct entry_s entry_t;

-[private] struct entry_s {
	char *string;
	void *data;
	-[ tree members for entry_t ]
}; 

-[private] typedef struct {
	entry_t *root;
} strtable_t;

-[ instantiate tree strtable_t : 
	scope= private,
	node type= entry_t,
	comparision= strcompare,
	prefix= strtree ]

#define getT   strtable_t *T= (strtable_t*) _T

-[public] void* strtab_find(strtab_t *_T, char *key) {
	getT;
	entry_t E;
	E.string= key;
	entry_t *R;
	R= strtree_find(T,&E);
	if (R)
		return R->data;
	else
		return 0;
}

-[private] int strcompare(entry_t *A,entry_t *B) {
	return strcmp(A->string,B->string);
}

-[ public] typedef struct { void* dummy; } strtab_t;

-[public] void strtab_insert(strtab_t *_T, char *key, void *data) {
	getT;
	entry_t *E;
	-[ptr/new E]
	E->string= strdup(key);
	E->data= data;
	strtree_insert(T,E);
}

-[ header "string.h" ]
-[ header "stdlib.h" ]

-[public] strtab_t *strtab_new() {
	strtable_t *T;
	-[ptr/new T]
	-[ptr/zero T]
	return (strtab_t*) T;
}

-[public] void*
strtab_destroy(strtab_t *_T) {
	getT;
	entry_t *E;
	E= strtree_destroy(T);
	if (E) {
		void *data= E->data;
		free(E->string);
		free(E);
		return data;
	} else 
		return 0;
}

-[public] void*
strtab_remove(strtab_t *_T,char *key) {
	getT;
	entry_t E,*R;
	E.string= key;
	R= strtree_remove(T, &E);
	if (R) {
		void *data= R->data;
		free(R->string);
		free(R);
		return data;
	} else
		return 0;
}


-[public] char *strtab_getleast(strtab_t *_T) {
	getT;
	return least(T->root);
}

-[private] char *least(entry_t *E) {
	if (!E)
		return 0;
	if (E->left)
		return least(E->left);
	return E->string;
}


void strtab_unused_functions_do_not_call() {
	strtree_init(0);
}
