

-[builtin function set, begin ]
	-[ arg _VARNAME ]
	-[ arg _VALUE ]
-[end]
{
	char *name, *value;
	name= get_single_string(var_get("_VARNAME"));
	if (!name)
		exit(fprintf(stderr,
			"jj: #set: empty variable name requested\n"));
	value= expand_variables_in_str(var_get("_VALUE"));
	name= mprintf("$%s",name);
	var_set(name, value);
	free(name);
}

-[ import strsplit function in private scope ]

-[builtin function local, begin ]
	-[ arg _VARNAMES ]
-[end]
{
	char *name;
	char **name_components;
	int i;
	name= var_get("_VARNAMES");
	name_components= strsplit( "[[:space:]]+",name);
	if (!name_components || !name_components[0])
		exit(fprintf(stderr,
			"jj: #local: empty variable name requested\n"));
	for(i=0;name_components[i];i++) {
		name= mprintf("$%s",name_components[i]);
		var_create(name, strdup(""));
		free(name);
	}
	free_string_array(name_components);
}

-[builtin function function, begin]
	-[ arg _NAME ]
	-[ arg _ARGS ]
	-[ arg _BODY ]
-[end]
{
	char *name, **args, *body;
	func_t *F;
	int i;

	name= get_single_string(var_get("_NAME"));
	if (!name)
		exit(fprintf(stderr,"jj: #function: empty name "
			"given for function\n"));
	args= split_string_into_array(var_get("_ARGS"));
	body= var_get("_BODY");
	
	name= mprintf("#%s",name);

	F= func_new(name);
	F->kind= fn_user;
	F->data= strdup(body);
	for(i=0;args[i];i++) {
		char *A;
		A= mprintf("$%s",args[i]);
		strlist_append(&F->args, A);
	}
	free_string_array(args);
}


-[private] char* get_single_string(char *str) {
	char **components;
	int i;
	char *R;
	components= strsplit( "[[:space:]]+",str);
	if (!components)
		return 0;
	if (!components[0]) {
		free(components);
		return 0;
	}
	R= components[0];
	for(i=1;components[i];i++)
		free(components[i]);
	free(components);
	return R;
}

-[private] char** split_string_into_array(char *str) {
	return strsplit( "[[:space:]]+",str);
}

-[private] void free_string_array(char **X) {
	int i;
	if (!X) return ;
	for(i=0;X[i];i++) free(X[i]);
	free(X);
}


-[builtin function strcat, begin]
	-[ arg _RESULT ]
	-[ arg _STR1 ]
	-[ arg _STR2 ]
-[end] {
	char *resultvar;
	char **str1_w, **str2_w;
	char *result;
	int L1,L2;
	int i,j;
	charbuffer_t *buffer;

	resultvar= get_single_string(var_get("_RESULT"));

	if (!resultvar)
		exit(fprintf(stderr,"jj: #strcat: result variable must "
			"not be null\n"));

	str1_w= split_string_into_array(
		expand_variables_in_str(var_get("_STR1")));
	str2_w= split_string_into_array(
		expand_variables_in_str(var_get("_STR2")));
	L1= string_array_len(str1_w);
	L2= string_array_len(str2_w);

	if (L1==0||L2==0) {
		char *z;
		if (L1==0) z= "_STR2";
		else z= "_STR1";
		result= expand_variables_in_str(var_get(z));
		goto end;
	}

	buffer= charbuffer_new();
	for(i=0;i<L1;i++)
		for(j=0;j<L2;j++) {
			if (i!=0||j!=0)
				charbuffer_append(buffer,' ');
			charbuffer_putstr(buffer,str1_w[i]);
			charbuffer_putstr(buffer,str2_w[j]);
		}
	charbuffer_append(buffer,0);

	result= charbuffer_collectda(buffer,&L1);
	free_string_array(str1_w);
	free_string_array(str2_w);
end:
	resultvar= mprintf("$%s",resultvar);
	var_set(resultvar, result);
	free(resultvar);
}


-[private] void charbuffer_putstr(charbuffer_t *B,char *str) {
	while(*str) {
		charbuffer_append(B,str[0]);
		str++;
	}
}

-[private] int string_array_len(char **x) {
	int i;
	if (!x) return 0;
	for(i=0;x[i];i++)
		;
	return i;
}

-[ builtin function foreach, begin ]
	-[ arg _VARNAME ]
	-[ arg _LIST ]
	-[ arg _BODY ]
-[ end ] {
	char *name;
	char **value;
	char *body;
	int i;
	name= get_single_string(var_get("_VARNAME"));
	if (!name)
		exit(fprintf(stderr,
			"jj: #foreach: empty variable name requested\n"));
	name= mprintf("$%s",name);
	value= split_string_into_array(
		expand_variables_in_str(var_get("_LIST")));
	body= var_get("_BODY");
	frame_enter();
	var_create(name, "");
	for(i=0;value[i];i++) {
		var_set(name, value[i]);
		ipret_block("<internal>",body);
	}
	free(name);
	frame_leave();
	free_string_array(value);
}

-[ builtin function include, begin ]
	-[ arg _FILENAME ]
-[ end ] { 
	char **fn;
	char *here;
	int i;
	here= strdup(var_get("$HERE"));
	fn= split_string_into_array(
		expand_variables_in_str(var_get("_FILENAME")));
	for(i=0;fn[i];i++) {
		ipret_file(here, fn[i]);
	}
	free_string_array(fn);
}

-[ builtin function replace_suffix, begin ]
	-[ arg _RESULT ]
	-[ arg _SOURCE ]
	-[ arg _NEWSFX ]
-[ end ]
{
	char *resultvar;
	char **source;
	char *newsfx;
	charbuffer_t *buffer;
	int i;

	resultvar=  get_single_string(var_get("_RESULT"));
	source= split_string_into_array(
		expand_variables_in_str(var_get("_SOURCE")));
	newsfx= get_single_string(var_get("_NEWSFX"));

	buffer= charbuffer_new();
	for(i=0;source[i];i++) {
		if (i) charbuffer_append(buffer,' ');
		charbuffer_putmem(buffer, source[i],
			find_suffix_begin(source[i])-source[i]);
		charbuffer_putstr(buffer, newsfx);
	}
	charbuffer_append(buffer,0);
	resultvar= mprintf("$%s",resultvar);
	var_set(resultvar, charbuffer_collectda(buffer,&i));
	free_string_array(source);
	free(resultvar);
}


-[private] char *find_suffix_begin(char *str) {
	char *ret;
	ret= 0;
	while(*str) {
		if (*str=='.') ret= str;
		str++;
	}
	if (ret) return ret;
	return str;
}

-[private] void charbuffer_putmem
	(charbuffer_t *buffer,char *mem,int len)
{
	int i;
	for(i=0;i<len;i++)
		charbuffer_append(buffer,mem[i]);
}

-[ builtin function nth, begin ]
	-[ arg _RESULT ]
	-[ arg _SOURCE ]
	-[ arg _INDEX ]
-[ end ]
{
	int L;
	char *resultvar,*index,**source;
	int W;

	resultvar= get_single_string(var_get("_RESULT"));
	source= split_string_into_array(
		expand_variables_in_str(var_get("_SOURCE")));
	index= get_single_string(
		expand_variables_in_str(var_get("_INDEX")));
	for(L=0;source[L];L++)
		;
	W= atoi(index);
	resultvar=mprintf("$%s",resultvar);
	if (W<0 || W>=L)
		var_set(resultvar,"");
	else
		var_set(resultvar, source[W]);
	free(resultvar);
	free_string_array(source);
}

