
-[ symbollist_t: public list of symbol_t* ]

-[public] typedef struct {
	symbollist_t *fields;
	char **triggers;
	char **outputs;
	char *message;
	char *action;
} command_t;


-[ builtin function command, begin ]
	-[ arg _BODY ]
-[ end ] {
	strlist_t *body;
	char *S;
	char **target;
	command_t *C;

	body= parse_strlist(var_get("_BODY"));

	-[ allocate object C ]
	C->fields= symbollist_new();

	target= 0;
	-[ list>for each S in *body, begin ]
		if (target) {
			*target= expand_variables_in_str(S);
			target= 0;
		} else {
			symbol_t *Y;
			-[ allocate object Y ]
			Y->name= S;
			target= &Y->value;
			symbollist_append(C->fields,Y);
		}
	-[ end ]

	while((S=strlist_destroy(body)))
		/* do not free, we stored them */ ;

	commandlist_append(commands, C);
}

-[ commandlist_t : public list of command_t* ]

-[ public ] commandlist_t* commands;

-[private] strlist_t *parse_strlist(char *input)
{
	strlist_t *R;
	int state= 0;
	int Z;
	char *t_begin;

	R= strlist_new();
	state= 0;

	while(1) {
	-[tuple switch state X *input, begin]
	-[case 0 X 0 ] goto end;
	-[case 0 X 127 ]
	-[caser 0 X 1 .. 32] input++; break;
	-[case 0 X LCURLY ] Z= 1; -[ token transition 'C']
	-[case 0 X RCURLY] 
		do_error("<internal>", input,  "unmatched curly brace");
		break;
	-[case 0 X '$' ] -[ token transition 'W' ]
	-[cased 0 ]  -[ token transition 'W']

	-[case 'C' X LCURLY ]  Z++; input++; break;
	-[case 'C' X RCURLY ]
		input++;
		Z--;
		if (Z!=0)
			break;
		t_begin++;
		strlist_append(R,  make_str(t_begin, input-t_begin-1));
		state=0;
		break;
	-[case 'C' X 0 ]
		do_error("<internal>", 0,"unexpected EOF in string");
		break;
	-[cased 'C' ] input++; break;

	-[case 'W' X 0 ]
	-[caser 'W' X 1 .. 32]
	-[case 'W' X RCURLY]
	-[case 'W' X LCURLY]
		state= 0;
		strlist_append(R,make_str(t_begin,input-t_begin));
		break;
	-[cased 'W'] input++; break;

	-[ dcase 0 ]
		do_error("<internal>", 0,"unexptected EOF in string");
	-[end]
	}

end:
	return R;
}

=[ token transition $STATE ]
	t_begin= input;
	input++;
	state= $STATE;
	break;

=[]

-[public] void cmd_init() {
	commands= commandlist_new();
}


-[public] void prepare_commands() {
	-[ list>for each C in *commands,begin]
		prepare_command(C);
		if (!C->action || !C->outputs) {
			fprintf(stderr,"jj: removing mysterious command"
				"with no action or output files\n");
			-[ remove this ]
			/* FIXME: free C also */
		}
	-[ end ]
}

-[private] void prepare_command(command_t *C)
{
	frame_enter();
	-[list> for each S in *(C->fields) , begin ]
		char *g;
		g= mprintf("$%s",S->name);
		var_create(g,S->value);
		free(g);
	-[end]

	C->message= get_simple_field( "$message" );
	C->action= get_simple_field( "$action" );
	replace_newlines_with_spaces(C->action);
	C->outputs= get_array_field( "$outputs" );
	C->triggers= get_array_field( "$triggers" );
	frame_leave();
}

-[private] void replace_newlines_with_spaces(char *X) {
	int i;
	for(i=0;X[i];i++)
		if ((X[i]>0 && X[i]<' ') || X[i]==127)
			X[i]= ' ';
}

-[private] char *get_simple_field(char *vname)
{
	return expand_variables_in_str(strdup(var_get(vname)));
}

-[private] char **get_array_field(char *vname)
{
	char *val;
	val= var_get(vname);
	if (!val)
		return 0;
	val= expand_variables_in_str(strdup(val));
	return split_string_into_array(val);
}

/* FIXME: consider what happens when you have
	-l$XX etc.. in this case, expand_var would
	really expand XX. do we want that? 
or consider:
	$XX/yy
	what happens then? */

-[ strtree_t : tree of char* , begin]
	-[ scope= public]
	-[ compare ]
		return strcmp(X1,X2);
-[ end ]


-[public] void run_commands() {
	int change;
	strtree_t *input_set;

	-[ allocate object input_set ]
	strtree_init(input_set);

	-[ list>for each C in *commands, begin ]
		int i;
		for(i=0;C->triggers[i];i++)
			strtree_insert(input_set,C->triggers[i]);
	-[ end ]
	
	-[ list>for each C in *commands, begin ]
		int i;
		for(i=0;C->outputs[i];i++)
			strtree_remove(input_set,C->outputs[i]);
	-[ end ]

	change= 1;
	while(change && !commandlist_isempty(commands)) {
		change= 0;
		-[ list>for each C in *commands, begin ]
			if (execute_command(input_set,C)) {
				-[ remove this ]
				change= 1;
			}
		-[ end ]
	}
}



-[private] int execute_command
	(strtree_t *input_set,command_t *C)
{
	int i;
	time_t input_mtime;
	time_t output_mtime;

	for(i=0;C->triggers[i];i++)
		if (!strtree_find(input_set,C->triggers[i]))
			return 0;

	input_mtime= 0;
	for(i=0;C->triggers[i];i++) {
		time_t t;
		t= get_modtime(C->triggers[i]);
		if (t>input_mtime)
			input_mtime= t;
	}

	for(i=0;C->outputs[i];i++) {
		time_t t;
		t= get_modtime_weak(C->outputs[i]);
		if (i==0 || t<output_mtime)
			output_mtime= t;
	}

	if (input_mtime<output_mtime)
		goto all_done;

	if (messages_only) {
		printf("%s\n",C->message ? C->message : C->action);
	} else {
		printf("%s\n",C->action);
	}

	if (system(C->action))
		exit(fprintf(stderr,"jj: build failed. exiting\n"));

	for(i=0;C->outputs[i];i++)
		get_modtime_force(C->outputs[i]);

all_done:
	for(i=0;C->outputs[i];i++)
		strtree_insert(input_set,C->outputs[i]);
	return 1;
}


-[ builtin function modify, begin]
	-[ arg _TARGETS ]
	-[ arg _BODY ]
-[ end ] {
	char **targets;
	strlist_t *body;

	targets= split_string_into_array(
			expand_variables_in_str(var_get("_TARGETS")));
	body= parse_strlist(var_get("_BODY"));

	-[ list>for each C in *commands, begin ]
		if (command_has_target(C,targets))
			modify_command(C, body);
	-[ end ]
}


-[private] int command_has_target
	(command_t *C,char **targets)
{
	int i,j;
	int R;
	prepare_command(C);

	R= 0;
	for(i=0;C->outputs[i];i++)
		for(j=0;targets[j];j++)
			if (!strcmp(C->outputs[i], targets[j]))
				{ R=1; break; }

	free(C->message);
	free(C->action);
	free_string_array(C->outputs);
	free_string_array(C->triggers);

	return R;
}

-[private] void modify_command
	(command_t *C,strlist_t *body)
{
	char *target;
	target= 0;
	-[list>for each S in *body, begin ]
		if (target) {
			modify_field(C->fields, target, S);
			target= 0;
		} else {
			target= S;
		}
	-[ end ]
}


-[private] void modify_field
	(symbollist_t *SL,char *field,char *value)
{
	int do_append= 0;
	symbol_t *L;
	if (field[0]=='+') {
		field++;
		do_append= 1;
	}
	-[ list> for each S in *SL, begin ]
		if (!strcmp(S->name,field)) {
			if (do_append && S->value && S->value[0]) {
					S->value= mprintf("%s %s",
						S->value, value);
			} else S->value= strdup(value);
			return ;
		}
	-[ end ]
	
	-[ allocate object L ]
	L->name= strdup(field);
	L->value= strdup(value);
	symbollist_append(SL,L);
}
