
-[public] typedef struct {
	char *name;
	char *value;
} symbol_t;

-[ symtab_t : tree of symbol_t* , begin ]
	-[ scope = public ]
	-[ compare ]
		return strcmp(X1->name,X2->name);
-[ end ]

-[ context_t : public stack of symtab_t* ]

-[private] context_t *context;

-[public] char *var_get(char *name)
{
	symbol_t *R;
	R= find_var(name);
	if (R)
		return R->value;
	return 0;
}

-[private] symbol_t* find_var(char *name) {
	symbol_t fake;
	int i;
	fake.name= name;
	for(i=context->nelts-1;i>=0;i--) {
		symbol_t *R;
		R= symtab_find(context->elts[i], &fake);
		if (R) return R;
	}
	return 0;
}

-[public] void var_set(char *name,char *val)
{
	symbol_t *R;
	R= find_var(name);
	if (R) {
		free(R->value);
		R->value= strdup(val);
	} else {
		-[ allocate object R ]
		R->name= strdup(name);
		R->value= strdup(val);
		symtab_insert(context->elts[0],R);
	}
}

-[public] void var_create(char *name,char *val) {
	symbol_t fake;
	fake.name= name;
	if (symtab_find(context->elts[context->nelts-1], &fake)) {
		var_set(name,val);
	} else {
		symbol_t *R;
		-[ allocate object R ]
		R->name= strdup(name);
		R->value= strdup(val);
		symtab_insert(context->elts[context->nelts-1],R);
	}
}

-[public] void frame_enter() {
	symtab_t *symtab;
	-[ allocate object symtab ]
	symtab_init(symtab);
	context_push(context, symtab);
}

-[public] void frame_leave() {
	symtab_t *symtab;
	symbol_t *S;
	symtab= context_pop(context);
	while((S=symtab_destroy(symtab))) {
		free(S->name);
		free(S->value);
		free(S);
	}
	free(symtab);
}


-[private] void var_init() {
	context= context_new();
	frame_enter();
}

